/* -*-Mode: C++;-*-
 * PRCS - The Project Revision Control System
 * Copyright (C) 1997  Josh MacDonald
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: projdesc.h 1.3.1.1.1.2.1.10.1.1.1.6.3.8.1.2 Thu, 20 Feb 1997 12:32:28 -0800 jmacd $
 */


#ifndef _PROJDESC_H_
#define _PROJDESC_H_

EXTERN const char *prj_lex_val;
EXTERN int         prj_lex_len;
EXTERN int         prj_lex_this_index;
EXTERN int         prj_lex_cur_index;
EXTERN FILE*       prj_input;
EXTERN void        prjsetbuf(const char* seg, size_t len);
EXTERN int         prjlex();
EXTERN int         prjinput(void* buf, int max_size);

enum PrjTokenType {
    PrjString,
    PrjName,
    PrjOpen,
    PrjClose,
    PrjNull,
    PrjBadString,
    PrjEof
};

#ifdef __cplusplus

#define foreach_fileent(name, project) \
    foreach(name, (project)->file_entries(), FileEntryPtrArray::ArrayIterator)

class PrjFields {
public:
    const char *name;
    PrVoidError(ProjectDescriptor:: *parse_func)(const Sexp* list);
    PrVoidError(ProjectDescriptor:: *read_func) (int offset);
    bool must_have;
};

class Estring : public Dstring {

public:
    Estring (ProjectDescriptor* project,
	     char* s,
	     size_t off,
	     size_t len,
	     EstringType type);
    Estring (ProjectDescriptor* project, size_t position); /* type = EsUnProtected */

    virtual bool is_dstring();

    void append_protected(const char*); /* To insert a NameLiteral into an EsUnProtected string */
    void append_string(const char*); /* To insert a StringLiteral into an EsUnProtected string */

    EstringType type() const;

    /* A special Dstring that remembers where it was in the project
     * file, allows itself to be modified, adds itself to the
     * descriptors list of strings, and will update the project file
     * when it needs to get written out.  Modify is called before the
     * string is modified.  the alloc field will be set to -1 until
     * it is actually modified. */
    virtual void modify();

    size_t offset()        const;
    size_t offset_length() const;

protected:

    size_t _offset;
    size_t _len;
    EstringType _type;
};

class ListMarker {
public:
    ListMarker (size_t start0, size_t stop0) :start(start0), stop(stop0) { }
    ListMarker () :start(0), stop(0) { }
    size_t start;
    size_t stop;
};

class MergeParentFile {
public:
    MergeParentFile (const char* common0, const char* selected0,
		     const char* working0, MergeAction action0)
	:common(common0), selected(selected0),
	 working(working0), action(action0) { }

    const char* common;
    const char* selected;
    const char* working;
    MergeAction action;
};

class MergeParentEntry {
public:
    MergeParentEntry (const char* s_version0, const char* ew_version0, bool complete0, int lineno0)
	:selected_version(s_version0),
	 eff_working_version(ew_version0),
	 complete(complete0),
	 lineno(lineno0),
	 working_table(NULL),
	 common_table(NULL),
	 selected_table(NULL) { }

    Dstring selected_version;
    Dstring eff_working_version;
    bool complete;
    int lineno;
    MergeParentTable *working_table;
    MergeParentTable *common_table;
    MergeParentTable *selected_table;
    MergeParentFilePtrArray all_files;
};

PrProjectDescriptorPtrError read_project_file(const char* project_full_name,
					      const char* file_name,
					      ProjectReadData flags);
PrProjectDescriptorPtrError read_project_file(const char* project_full_name,
					      const char* file_name,
					      FILE* pipe,
					      ProjectReadData flags);
ProjectDescriptor* fake_project_descriptor();

class ProjectDescriptor {

    friend PrProjectDescriptorPtrError read_project_file(const char* project_full_name,
							 const char* filename,
							 ProjectReadData flags);

    friend PrProjectDescriptorPtrError read_project_file(const char* project_full_name,
							 const char* file_name,
							 FILE* pipe,
							 ProjectReadData flags);

    friend ProjectDescriptor* fake_project_descriptor();

    friend class Estring;

    friend PrBoolError setkeys_internal(const char* input_buffer0,
					int input_buffer_len0,
					ostream* os,
					MemorySegment* seg,
					FileEntry *fe,
					ProjectDescriptor *project,
					SetkeysAction action);
public:

    ~ProjectDescriptor();

    PrVoidError write_project_file(const char* filename);
    void append_files_data (const char*); /* unprotected, for comments and such. */
    void append_file (const char* working_path,
		      bool keywords);
    void append_file (FileEntry* fe);
    void append_link (const char* working_path,
		      const char* link_name);
    void append_directory (const char* working_path);
    void delete_file (FileEntry* fe);

    Dstring* project_version_name()  const;
    Dstring* project_version_major() const;
    Dstring* project_version_minor() const;

    Dstring* parent_version_name()  const;
    Dstring* parent_version_major() const;
    Dstring* parent_version_minor() const;

    Dstring* project_description() const;
    Dstring* version_log()         const;
    Dstring* new_version_log()     const;
    Dstring* checkin_time()        const;
    Dstring* checkin_login()       const;
    Dstring* created_by_major()    const;
    Dstring* created_by_minor()    const;
    Dstring* created_by_micro()    const;

    Dstring* descends_from_name()  const;
    Dstring* descends_from_major() const;
    Dstring* descends_from_minor() const;

    /* merge_continuing instructs the project descriptor that future
     * merge_notify and has_been_merged requests refer to the named
     * project. */
    bool merge_continuing(ProjectVersionData* selected, ProjectVersionData *eff_working);
    void merge_notify_incomplete();
    void merge_notify_complete();
    void merge_notify(FileEntry* w, FileEntry* c, FileEntry* s, MergeAction ma) const;
    bool has_been_merged(FileEntry* w, FileEntry* c, FileEntry* s) const;
    void merge_finish();
    PrVoidError adjust_merge_parents();
    MergeParentEntryPtrArray *new_merge_parents() const;
    PrVoidError check_version_name (const char* name, int line) const;

    DstringPtrArray* populate_ignore() const;

    void read_quick_elim();
    void quick_elim_unmodified();
    void quick_elim_update();
    void quick_elim_update_fe(FileEntry* fe);

    PrVoidError update_repository();

    ProjectVersionData* project_version_data() const;

    FileEntryPtrArray* file_entries();

    PrProjectDescriptorPtrError resolve_checkin_version(const char* maj);
    PrVoidError checkin_project_file();

    void set_descends_from(const char* major, int minor);

    PrVoidError init_repository_entry(const char* name, bool lock, bool create);
    RepEntry* repository_entry() const;
    void repository_entry(RepEntry* rep);

    const char* project_full_name() const;
    const char* project_name()      const;
    const char* project_file_path() const;
    const char* project_aux_path()  const;
    const char* project_path()      const;
    const char* full_version()      const;

private:

    KeywordTable* project_keywords (FileEntry* fe, bool setting);

    PrVoidError created_by_prj_entry_func        (const Sexp* s);
    PrVoidError project_ver_prj_entry_func       (const Sexp* s);
    PrVoidError parent_ver_prj_entry_func        (const Sexp* s);
    PrVoidError project_desc_prj_entry_func      (const Sexp* s);
    PrVoidError version_log_prj_entry_func       (const Sexp* s);
    PrVoidError new_version_log_prj_entry_func   (const Sexp* s);
    PrVoidError checkin_time_prj_entry_func      (const Sexp* s);
    PrVoidError checkin_login_prj_entry_func     (const Sexp* s);
    PrVoidError descends_from_prj_entry_func     (const Sexp* s);
    PrVoidError populate_ignore_prj_entry_func   (const Sexp* s);
    PrVoidError project_keywords_prj_entry_func  (const Sexp* s);

    /* These 3 entries in the project file grow with project
     * file count, so reading the entire list is a waste of
     * memory. */
    PrVoidError files_prj_entry_read             (int offset);
    PrVoidError merge_parents_prj_entry_read     (int offset);
    PrVoidError new_merge_parents_prj_entry_read (int offset);

    /* Read helper functions. */
    PrVoidError read_merge_parents               (MergeParentEntryPtrArray* array,
						  ListMarker* finish);
    PrVoidError read_merge_parents_entry         (MergeParentEntryPtrArray* array);
    PrVoidError read_merge_parents_file_name     (const Sexp* s, Estring** fill);
    PrVoidError read_merge_parents_insert_file   (MergeParentTable* table,
						  Estring* filename,
						  const Sexp* s,
						  MergeParentFile* file);
    PrVoidError individual_insert_fileent        (const Sexp* s);
    PrVoidError verify_project_file              ();
    PrVoidError bad_project_file                 (const Sexp* s, const char* message);

    /* Friendly constructors. */
    PrVoidError init_from_file(const char* project_full_name, const char* filename,
			       FILE* infile, ProjectReadData flags);
    void init_fake_project();
    ProjectDescriptor();

    /* Illegal. */
    ProjectDescriptor(const ProjectDescriptor&);
    operator=(const ProjectDescriptor&);

    /* Parsing. */
    int prj_lookup_hash (const char *str, int len);
    const PrjFields *prj_lookup_func (const char *str, int len);
    PrConstSexpPtrError read_list(int offset);
    PrConstSexpPtrError read_list_token(PrjTokenType type, int offset);
    PrConstSexpPtrError read_list_elt();
    PrConstSexpPtrError read_new_estring(PrjTokenType type);
    PrVoidError read_list_checked(size_t offset);
    PrVoidError read_parse_bad_token(PrjTokenType type);
    PrVoidError parse_prj_file(FILE* infile);
    void register_estring(Estring*, char missing);

    /* Writing. */
    void really_write_project_file (ostream& os);
    void write_merge_parents (const char* name, MergeParentEntryPtrArray *table, ostream& os);

    /* Cleanup */
    void delete_merge_parents(MergeParentEntryPtrArray* parents);

    /* Static data. */
    static const int       _prj_entries;
    static const PrjFields _pftable[];
    static bool            _prj_entry_found[];

    /* Parsing data. */
    const char* _prj_source_name;
    ProjectReadData _read_flags;

    /* QE buffer, if it was available. */
    QuickElim* _quick_elim;

    /* Names */
    const char* _project_name;
    Dstring  _full_version;
    Dstring  _project_full_name;
    Dstring  _project_file_path;
    Dstring  _project_path;
    Dstring  _project_aux_path;
    Dstring* _new_rcs_version;

    /* Populate */
    DstringPtrArray* _populate_ignore;

    /* Insertion points and the original buffer. */
    MemorySegment   *_segment;
    Dstring*         _nullified_chars;
    DstringPtrArray *_all_estrings;
    Estring         *_files_insertion_point;
    Estring         *_project_version_point;
    Estring         *_end_of_buffer_point;
    ListMarker       _merge_parents_marker;
    ListMarker       _new_merge_parents_marker;
    ListMarker       _descends_from_marker;
    ConstListMarkerPtrArray *_deleted_markers;

    /* Repository data. */
    RepEntry*           _repository_entry;
    FileEntryPtrArray*  _file_entries;
    ProjectVersionData* _project_version_data;

    /* Project file strings. */
    Dstring*      _project_version_name;
    Dstring*      _project_version_major;
    Dstring*      _project_version_minor;

    Dstring*      _parent_version_name;
    Dstring*      _parent_version_major;
    Dstring*      _parent_version_minor;

    Dstring*      _project_description;
    Dstring*      _version_log;
    Dstring*      _new_version_log;
    Dstring*      _checkin_time;
    Dstring*      _checkin_login;

    Dstring*      _created_by_major;
    Dstring*      _created_by_minor;
    Dstring*      _created_by_micro;

    Dstring*      _descends_from_name;
    Dstring*      _descends_from_major;
    Dstring*      _descends_from_minor;

    MergeParentEntryPtrArray *_merge_parents;
    MergeParentEntryPtrArray *_new_merge_parents;

    /* Keyword stuff */
    KeywordTable* _project_keywords;
    Dstring* _keyword_id;
    Dstring* _keyword_pheader;
    Dstring* _keyword_pversion;

    /* Current Merge Data */
    MergeParentEntry *_merge_entry;

    const char* _merge_selected_major;
    const char* _merge_selected_minor;

    bool _append_merge_parents;

    bool _merge_complete;
    bool _merge_incomplete;
};

#endif

#endif /* PROJDESC_H */
